local addonInfo, privateVars = ...

---------- init namespace ---------

if not EnKai then EnKai = {} end

if not privateVars.uiFunctions then privateVars.uiFunctions = {} end

local uiFunctions   = privateVars.uiFunctions
local internal      = privateVars.internal
local mapData       = privateVars.mapData

---------- addon internal function block ---------

local function _uiMapElementTexture(name, parent)

  local thisMapData = nil
  local parentMap = nil
  local zoom = 1
  local tooltipTitle = nil
  local tooltipLines = nil
  local maxZoom = 4
  local waypoint = nil
  local effectName = nil
  local tooltip = false
  
  local coordX, coordY

	--if EnKai.internal.checkEvents (name, true) == false then return nil end
	
  local mapElement = EnKai.uiCreateFrame("nkTexture", name .. ".texture", parent)
  mapElement:SetLayer(9)
  
  function mapElement:SetParentMap(newParentMap) parentMap = newParentMap end

  function mapElement:SetType(elementType)
    thisMapData = mapData.mapElements[elementType]
    local addon = "Rift"
    if thisMapData.addon ~= nil then addon = thisMapData.addon end
    mapElement:SetTextureAsync (addon, thisMapData.path)
    
    if thisMapData.layer ~= nil then mapElement:SetLayer(thisMapData.layer) end
        
  end
  
  function mapElement:SetToolTip(title, newDesc)
  
    if title == nil and newDesc == nil then return end
    
    local descStart
    
    if title ~= nil then    
      tooltipTitle = title
      if newDesc ~= nil then descStart = 1 end
    else 
      tooltipTitle, descStart = newDesc[1], 2
    end
    
    tooltipLines = {}
    
    if descStart ~= nil then
      for idx = descStart, #newDesc, 1 do
        table.insert (tooltipLines, { text = newDesc[idx] })
      end
    end
    
  end

  function mapElement:SetCoord(x, y)
  
    if x ~= nil then coordX = x end
    if y ~= nil then coordY = y end
    
    local mapInfo = parentMap:GetMapInfo()
    
    local xP = 1 / (mapInfo.x2 - mapInfo.x1) * (coordX - mapInfo.x1)
    local yP = 1 /  (mapInfo.y2 - mapInfo.y1) * (coordY - mapInfo.y1) 
    
    local thisX, thisY
    
    thisX = (parentMap:GetMap():GetWidth() * xP) - (mapElement:GetWidth() / 2)
    thisY = (parentMap:GetMap():GetHeight() * yP) - (mapElement:GetWidth() / 2)
    
    mapElement:SetPoint("TOPLEFT", parentMap:GetMap(), "TOPLEFT", thisX, thisY)
    
  end
  
  function mapElement:SetZoom(newZoom)
    
    local factor = thisMapData.factor or 1
    
    if thisMapData.minZoom ~= nil then 
      if thisMapData.minZoom > newZoom then     
        mapElement:SetHeight((thisMapData.width * factor) / thisMapData.minZoom * newZoom)
        mapElement:SetWidth((thisMapData.height * factor) / thisMapData.minZoom * newZoom) 
      else
        mapElement:SetHeight(thisMapData.width * factor)
        mapElement:SetWidth(thisMapData.height * factor)
      end
    else
      mapElement:SetHeight((thisMapData.width * factor) / maxZoom * newZoom)
      mapElement:SetWidth((thisMapData.height * factor) / maxZoom * newZoom)
    end

    zoom = newZoom
    
    if coordX ~= nil and coordY ~= nil then mapElement:SetCoord() end
    
  end

  function mapElement:GetCoord() return coordX, coordY end
  function mapElement:GetTooltip() return tooltip end
  
  ----- EVENTS
  
  mapElement:EventAttach(Event.UI.Input.Mouse.Cursor.In, function ()
    if tooltipTitle == nil then return end
    parentMap:GetTooltip():SetPoint("TOPLEFT", mapElement, "CENTER")
    
    parentMap:GetTooltip():SetTitle(tooltipTitle)
    parentMap:GetTooltip():SetLines(tooltipLines)
    parentMap:GetTooltip():SetVisible(true)
    tooltip = true
  end, mapElement:GetName() .. ".Cursor.In")
  
  mapElement:EventAttach(Event.UI.Input.Mouse.Cursor.Out, function ()
    parentMap:GetTooltip():SetVisible(false)
    tooltip = false
  end, mapElement:GetName() .. ".Cursor.Out")
  
  return mapElement
  
end

uiFunctions.NKMAPELEMENTTEXTURE = _uiMapElementTexture