local addonInfo, privateVars = ...

---------- init namespace ---------

if not EnKai then EnKai = {} end

if not privateVars.uiFunctions then privateVars.uiFunctions = {} end
if not privateVars.uiNames then privateVars.uiNames = {} end
if privateVars.uiContext == nil then privateVars.uiContext = UI.CreateContext("EnKai.ui") end

local data       = privateVars.data
local internal   = privateVars.internal
local uiFunctions= privateVars.uiFunctions
local uiNames    = privateVars.uiNames

local uiContext   = privateVars.uiContext

---------- init variables --------- 

data.frameCount = 0
data.canvasCount = 0
data.textCount = 0
data.textureCount = 0

---------- init local variables ---------

local _gc = {}
local _freeElements = {}

---------- local function block ---------

local function _recycleElement (element)
	
	element:SetVisible(false)
	element:ClearAll()
	element:SetBackgroundColor(0,0,0,0)
	element:SetStrata('main')
	element:SetLayer(0)
	element:SetMouseMasking('full')
	
	if element:GetMouseoverUnit() ~= nil then element:SetMouseoverUnit(nil) end
	
	--element:SetSecureMode("normal")
	
	for k, v in pairs (element:GetEvents()) do
	  element:EventDetach(k, nil, v.label, v.priority, v.owner)
	end
	
	element:_recycle()
	
end

local function _setInsecure (element)

  element:SetSecureMode("normal")

end

---------- addon internal function block ---------

-- the below is a prototype to frame resuing
-- missing SetName() on frames to fully build this

function internal.uiAddToGarbageCollector (frameType, element)

  local checkFrameType = string.upper(frameType) 

  if _gc[checkFrameType] == nil then _gc[checkFrameType] = {} end
  if _gc[checkFrameType].normal == nil then _gc[checkFrameType].normal = {} end
  if _gc[checkFrameType].restricted == nil then _gc[checkFrameType].restricted = {} end
  
  table.insert(_gc[checkFrameType][element:GetSecureMode()], element) 
  if Inspect.System.Secure() == false or element:GetSecureMode() == 'normal' then element:SetVisible(false) end
  
  EnKai.eventHandlers["EnKai.internal"]["gcChanged"]()
  
end  

function internal.uiGarbageCollector ()

	local secure = Inspect.System.Secure()
	local flag = false
	local restrictedFailed = false

	for elementType, secureModes in pairs(_gc) do
	
		if secure == false and #_gc[elementType].restricted > 0 then
			for idx = 1, #_gc[elementType].restricted, 1 do
			
			  if _gc[elementType].restricted[idx] ~= false then
			
  				local element = _gc[elementType].restricted[idx]
  				
  				local err = pcall (_setInsecure, element)
          if err == true then -- no error
            flag = true
          
    				uiNames[elementType][element:GetRealName()] = ""
  	   			_recycleElement(element)
  							
  				  if _freeElements[elementType] == nil then _freeElements[elementType] = {} end
  				  table.insert(_freeElements[elementType], element)
  				  _gc[elementType].restricted[idx] = false
  				else
  				  restrictedFailed = true
  				end
  			end
				
			end
		  
		  if restrictedFailed == false then _gc[elementType].restricted = {} end
		end
		
		for idx = 1, #_gc[elementType].normal, 1 do
			flag = true
			local element = _gc[elementType].normal[idx]
			uiNames[elementType][element:GetRealName()] = ""
			_recycleElement(element)
			
			if _freeElements[elementType] == nil then _freeElements[elementType] = {} end
			table.insert(_freeElements[elementType], element)
		end
		
		_gc[elementType].normal = {}
	end
	
	if flag == true then EnKai.eventHandlers["EnKai.internal"]["gcChanged"]() end

end

---------- library public function block ---------

function EnKai.uiCreateFrame (frameType, name, parent)

  if frameType == nil or name == nil or parent == nil then
    EnKai.tools.error.display (addonInfo.identifier, string.format("EnKai.uiCreateFrame - invalid number of parameters\nexpecting: type of frame (string), name of frame (string), parent of frame (object)\nreceived: %s, %s, %s", frameType, name, parent))
    return
  end

  local uiObject = nil

  local checkFrameType = string.upper(frameType) 

  if _freeElements[checkFrameType] ~= nil and #_freeElements[checkFrameType] > 0 then
    if EnKai.internal.checkEvents (name, true) == false then return nil end
    uiObject = _freeElements[checkFrameType][1]    
    uiObject:SetParent(parent)
    
    if uiNames[checkFrameType] == nil then uiNames[checkFrameType] = {} end
        
    uiNames[checkFrameType][uiObject:GetName()] = name
    uiObject:SetVisible(true)
    
    table.remove(_freeElements[checkFrameType], 1)
  else
    local func = uiFunctions[checkFrameType]
    if func == nil then
      EnKai.tools.error.display (addonInfo.identifier, string.format("EnKai.uiCreateFrame - unknown frame type [%s]", frameType))
    else
      uiObject = func(name, parent)
    end
  end

  return uiObject

end

function EnKai.getGcCount() return #_gc end
function EnKai.getFreeCount() return #_freeElements end