local addonInfo, privateVars = ...

privateVars.internal = {}

privateVars.uiElements = {}
privateVars.uiElements.context = UI.CreateContext("nkTextureViewer")
privateVars.uiElements.context:SetStrata("dialog")



function privateVars.internal.loadTextures(prefix)

  local hex = "0123456789ABCDEF"
  local hexValue
  local path
  local maxCounter = 5000
  local counter = 0
  local step = 25
  
  local texture = EnKai.uiCreateFrame("nkTexture", "testTexture", privateVars.uiElements.context)
  
  local getMainMapTextures = coroutine.create(
    function ()
      for idx = 1, maxCounter, step do
        for idx2 = idx, idx+(step-1), 1 do
        
          local hex = string.format("%X", idx2)
          path = prefix .. hex .. ".dds"
          if nkTVTextures[path] == nil then
            texture:ClearAll()
            texture:SetTexture("Rift", path)
            if texture:GetWidth() ~= 57 and texture:GetHeight() ~= 36 then
              counter = counter + 1
              nkTVTextures[path] = { width = texture:GetWidth(), height = texture:GetHeight(), recorded = Inspect.Time.Server()}
            end  
          end
        end
        
        --print (idx + (step-1) .. " done")
                  
        coroutine.yield(idx + (step-1))
      end
    end
  )

  local callBack = function ()
    print ("got " .. counter .. " new textures")
    texture:destroy()
    if privateVars.uiElements.viewer ~= nil then 
      privateVars.uiElements.viewer:PopulateGrid()
      privateVars.uiElements.viewer:SetVisible(true)
    else
      privateVars.internal.commandHandler ()
    end
  end
  
  EnKai.coroutines.add ({ func = getMainMapTextures, counter = maxCounter, active = true, callBack = callBack })
  

end

function privateVars.internal.main(_, addon)

  if addon == addonInfo.identifier then
  
    Command.Event.Attach(Event.System.Texture, privateVars.internal.recordTextures, "nkTextureViewer.System.Texture")
    Command.System.Texture.Record()
    
    table.insert(Command.Slash.Register("nkTV"), {privateVars.internal.commandHandler, "nkTextureViewer", "ui"})
    
    --privateVars.internal.viewer ()
   
  end
  
end

function privateVars.internal.settingsHandler(_, addon) 

  if addon == addonInfo.identifier then
  
    if nkTVTextures == nil then nkTVTextures = {} end 
    
  end
          
end

Command.Event.Attach(Event.Addon.Load.End, privateVars.internal.main, "nkTextureViewer.Addon.Load.End")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, privateVars.internal.settingsHandler, "nkTextureViewer.SavedVariables.Load.End")

function privateVars.internal.recordTextures (_, textures)
  
  local texture = EnKai.uiCreateFrame("nkTexture", "testTexture", privateVars.uiElements.context)
  
  for k, v in pairs(textures) do
    if nkTVTextures[k] == nil then
      texture:ClearAll()
      texture:SetTexture("Rift", k)
      if texture:GetWidth() ~= 57 and texture:GetHeight() ~= 36 then
        nkTVTextures[k] = { width = texture:GetWidth(), height = texture:GetHeight(), recorded = Inspect.Time.Server()}
      end 
    end
  end

end

function privateVars.internal.commandHandler (commandline)

  if commandline ~= nil and string.find(commandline, "load") ~= nil then
    local prefix = string.match(commandline, "load (.+)")
    print ("Searching for textures with prefix: " ..prefix)
    privateVars.internal.loadTextures(prefix)
  else
    if privateVars.uiElements.viewer == nil then
      privateVars.uiElements.viewer = privateVars.internal.viewer()
    elseif privateVars.uiElements.viewer:GetVisible() == true then
      privateVars.uiElements.viewer:SetVisible(false)
    else
      privateVars.uiElements.viewer:SetVisible(true)
    end
  end

end

function privateVars.internal.viewer ()

  local name = "nkTextureViewer"
  local gridRows = 25

  local viewer = EnKai.uiCreateFrame("nkWindowElement", name, privateVars.uiElements.context)
  viewer:SetPoint("TOPLEFT", UIParent, "TOPLEFT", 500, 300)
  viewer:SetWidth(840)
  viewer:SetHeight(530)
  viewer:SetTitle("Naifu's Texture Viewer")
  
  local searchLabel = EnKai.uiCreateFrame("nkText", name .. ".searchLabel", viewer:GetContent())
  searchLabel:SetPoint("TOPLEFT", viewer:GetContent(), "TOPLEFT", 0, 5)
  searchLabel:SetText("Search: ")
  
  local searchField = EnKai.uiCreateFrame("nkTextField", name .. ".searchField", viewer:GetContent())
  searchField:SetPoint("CENTERLEFT", searchLabel, "CENTERRIGHT", 5, 0)
  searchField:SetWidth(200)
  
  local cancelButton = EnKai.uiCreateFrame("nkTexture", name .. ".cancelButton", viewer:GetContent())
  cancelButton:SetPoint("CENTERLEFT", searchField, "CENTERRIGHT", 5, 0)
  cancelButton:SetWidth(12)
  cancelButton:SetHeight(12)
  cancelButton:SetTextureAsync("EnKai", "gfx/iconCancel.png")
  
  local sortRadio = EnKai.uiCreateFrame("nkRadioButton", name .. ".sortRadio", viewer:GetContent())
  sortRadio:SetPoint("CENTERLEFT", cancelButton, "CENTERRIGHT", 50, 0)
  sortRadio:SetWidth(200)
  sortRadio:SetLabelWidth(75)
  sortRadio:SetText("Sort by")
  sortRadio:SetSelection({{ value = "name", label = "name"}, { value = "date", label = "date"}})
  sortRadio:SetSelectedValue("name")
  sortRadio:SetColor(1,1,1,1)
  
  local filterIcons = EnKai.uiCreateFrame("nkCheckbox", name .. '.filterIcons', viewer:GetContent())  
  filterIcons:SetPoint("CENTERLEFT", sortRadio, "CENTERRIGHT", 50, 0)
  filterIcons:SetLabelWidth(100)
  filterIcons:SetText("filter icons")
  filterIcons:SetChecked(false)
  filterIcons:SetLabelInFront(true)
  filterIcons:SetColor(1,1,1,1)
  
  local testField = EnKai.uiCreateFrame("nkTextField", name .. ".testField", viewer:GetContent())
  testField:SetPoint("BOTTOMRIGHT", viewer:GetContent(), "BOTTOMRIGHT", -5, -5)
  testField:SetWidth(200)
  
  local testLabel = EnKai.uiCreateFrame("nkText", name .. ".testLabel", viewer:GetContent())
  testLabel:SetPoint("CENTERRIGHT", testField, "CENTERLEFT", -5, 0)
  testLabel:SetText("Test: ")
  
  local grid = EnKai.uiCreateFrame("nkGrid", name .. ".grid", viewer:GetContent())
  grid:SetHeaderHeight(0)
  grid:SetPoint("TOPLEFT", viewer:GetContent(), "TOPLEFT", 0, 30)
  grid:SetBorderColor(0.086, 0.235, 0.392, 1)
  grid:SetBodyColor(0.153, 0.314, 0.490, 1)
  grid:SetTransparentHeader()
  grid:SetLayer(2)
  
  local frame = EnKai.uiCreateFrame("nkFrame", name .. ".frame", viewer:GetContent())
  frame:SetPoint("TOPLEFT", grid, "TOPRIGHT", 10, 0)
  frame:SetWidth(475)
  frame:SetHeight(475)
  
  local texture = EnKai.uiCreateFrame("nkTexture", name .. ".texture", frame)
  texture:SetPoint("CENTER", frame, "CENTER")
  
  local nameLabel = EnKai.uiCreateFrame("nkText", name .. ".nameLabel", frame)
  nameLabel:SetPoint("TOPRIGHT", frame, "TOPRIGHT")
  
  local sizeLabel = EnKai.uiCreateFrame("nkText", name .. ".sizeLabel", frame)
  sizeLabel:SetPoint("TOPRIGHT", nameLabel, "BOTTOMRIGHT")
  
  local cols = {  { align = 'left', width = 250 }, { align = 'center', width = 50 }, { align = 'center', width = 50 } }
  local slider 
  
  grid:SetVisible(false)  
  grid:Layout (cols, gridRows)
  
  function viewer:ShowTexture(row)
    if grid:GetKey(row) == nil then return end
    viewer:SetTexture(grid:GetKey(row))    
  end  
  
  function viewer:SetTexture(path)
    texture:ClearAll()
    texture:SetPoint("CENTER", frame, "CENTER")
    texture:SetTexture("Rift", path)
    
    local width, height = texture:GetWidth(), texture:GetHeight()
    
    nameLabel:SetText(path)
    sizeLabel:SetText(string.format("%d / %d", width, height))    
    
    if width > height then
      if (width > 475) then
        local scale = 1 / width * 475
        texture:SetWidth(475)      
        texture:SetHeight(height * scale)
        texture:SetTexture("Rift", path)
      end  
    else
      if (height > 475) then
      local scale = 1 / height * 475
      texture:SetHeight(475)      
      texture:SetWidth(width * scale)
      texture:SetTexture("Rift", path)
      end
    end
  end
  
  cancelButton:EventAttach(Event.UI.Input.Mouse.Left.Click, function()
    grid:filter ("", 1, true)
  end, name .. ".UI.Input.Mouse.Left.Click")
  
  Command.Event.Attach(EnKai.events[name .. '.sortRadio'].RadiobuttonChanged, function (_, newValue)
    
    viewer:FilterIcons (filterIcons:GetChecked()) 
    --viewer:PopulateGrid(newValue)
    
  end, name .. '.sortRadio.RadiobuttonChanged')
  
  Command.Event.Attach(EnKai.events[name .. '.filterIcons'].CheckboxChanged, function (_, newValue)    
    viewer:FilterIcons (newValue)     
  end, name .. '.filterIcons.CheckboxChanged')
  
  Command.Event.Attach(EnKai.events[name .. '.searchField'].TextfieldChanged, function ()
    local text = searchField:GetText()
    
    if text == "" then
      grid:filter ("", nil, true)
    else
      grid:filter (text, 1, false)
    end
    
  end, name .. '.searchField.TextfieldChanged')
  
  Command.Event.Attach(EnKai.events[name .. '.testField'].TextfieldChanged, function ()
    local text = testField:GetText()
    viewer:SetTexture(text)
  end, name .. '.testField.TextfieldChanged')
                
  Command.Event.Attach(EnKai.events[name .. '.grid'].GridFinished, function ()
    viewer:PopulateGrid() 
    grid:SetVisible(true)
  end, name .. '.grid.GridFinished')
  
  Command.Event.Attach(EnKai.events[name .. '.grid'].WheelForward, function (_, rowPos)     
    slider:AdjustValue (rowPos)
    viewer:ShowTexture(grid:GetHighlightedRow())
  end, name .. '.grid.Grid.WheelForward')
  
  Command.Event.Attach(EnKai.events[name .. '.grid'].WheelBack, function (_, rowPos)
    slider:AdjustValue (rowPos)
    viewer:ShowTexture(grid:GetHighlightedRow())
  end, name .. '.grid.Grid.WheelBack')
  
  Command.Event.Attach(EnKai.events[name .. '.grid'].LeftClick, function (_, row)
    viewer:ShowTexture(row)
  end, name .. '.grid.LeftClick')   
  
  -- grid slider
  
  slider = EnKai.uiCreateFrame("nkScrollbox", name .. ".slider", viewer:GetContent()) 
  
  slider:SetPoint("TOPLEFT", grid, "TOPRIGHT")
  slider:SetHeight(grid:GetHeight())
  slider:SetRange(1, 1)
  slider:SetVisible(false)
  slider:SetLayer(2)
  slider:SetColor(1, 1, 1, 1)
  
  Command.Event.Attach(EnKai.events[name .. ".slider"].ScrollboxChanged, function ()    
    grid:SetRowPos(math.floor(slider:GetValue('value')), true)
  end, name .. '.slider.ScrollboxChanged')  
  
  function viewer:PopulateGrid (sort, textureList)
  
    local gridValues = {}
    
    local sourceList = textureList or nkTVTextures

    if sort == "date" then
      local tempValues = {}
      for key, details in pairs(sourceList) do
        table.insert(tempValues, {key, details.recorded})
      end
      
      table.sort (tempValues , function (v1, v2) return v1[2] > v2[2] end )
      
      for idx = 1, #tempValues, 1 do
        local key = tempValues[idx][1]
        local details = sourceList[key]
        local temp = os.date("*t", details.recorded)
        table.insert (gridValues, {{ key = key, value = key}, { value = details.width }, { value = details.height }  })
      end
      
    else
      for key, details in pairs(sourceList) do
        local temp = os.date("*t", details.recorded)
        table.insert (gridValues, {{ key = key, value = key}, { value = details.width }, { value = details.height }  })
      end
    
      table.sort (gridValues, function (v1, v2) return string.upper(v1[1].value) < string.upper(v2[1].value) end )
    end
    
    grid:SetRowPos(1, false)
    grid:SetCellValues(gridValues)

    if (#gridValues-gridRows) > 0 then    
      slider:SetRange (1, #gridValues-gridRows+1)
      slider:AdjustValue(1)
      slider:SetVisible(true)     
    else
      slider:SetVisible(false)
    end
    
    slider:AdjustValue(1)
    
  end
  
  function viewer:FilterIcons (flag)
  
    if flag == false then
      viewer:PopulateGrid (sortRadio:GetSelectedValue())
      return
    end
    
    local tempList = {}
  
    for key, details in pairs(nkTVTextures) do
      if details.width ~= 48 then
        tempList[key] = details
      end
    end
    
    viewer:PopulateGrid (sortRadio:GetSelectedValue(), tempList)
    
  end
  
  
  return viewer

end